@@ -21,11 +21,11 @@ from pathlib import Path
 from typing import Any
 
 import pytest
-from utils import load_config, template
 
-from yaecs import Configuration
-from yaecs.yaecs_utils import compare_string_pattern
+from unittests.config.utils import load_config, template
+from yaecs import Configuration, Experiment, Priority, assign_order, assign_yaml_tag
 from yaecs.user_utils import make_config
+from yaecs.yaecs_utils import compare_string_pattern
 
 
 def check_integrity(config, p_1: Any = 0.1, p_2: Any = 2.0, p_3: Any = 30.0,
@@ -133,6 +133,9 @@ def test_get_dict(yaml_default):
     assert config['save'] == 'test'
     assert isinstance(config['subconfig1'], Configuration)
     assert isinstance(config['subconfig2'], Configuration)
+    config = make_config({"a": 1}, post_processing_dict={"a": lambda x: x + 1})
+    assert config.get_dict(pre_post_processing_values=True) == {"a": 1}
+    assert config.get_dict(pre_post_processing_values=False) == {"a": 2}
 
 
 def test_iter(yaml_default):
@@ -230,7 +233,7 @@ def test_merge_from_command_line(caplog, yaml_default, yaml_experiment):
 
     def mcl(cfg, string):
         # pylint: disable=protected-access
-        to_merge = cfg._gather_command_line_dict(string_to_merge=string)
+        to_merge = cfg._gather_command_line_dict(to_merge=string)
         if to_merge:
             logging.getLogger("yaecs.config").info(f"Merging from command line : {to_merge}")
             cfg._merge(to_merge)
@@ -266,21 +269,17 @@ def test_merge_from_command_line(caplog, yaml_default, yaml_experiment):
     check_integrity(config_2, 1, 0.6, p_4="test test")
     with caplog.at_level(logging.WARNING):
         logging.getLogger("yaecs").propagate = True
-        mcl(config_2, "--param1 2 --*param2=none --*param3=none !str "
-                      "--*param4= '[ 1!int  ,0.5 !float, {string:\\'"
-                      "[as !str}!dict]' !list")
+        mcl(config_2, "--param1 2 --*param2=null --*param3=\\\"null\\\" "
+                      "--*param4= [ 1  ,0.5 , {string: \\\"\\'[as \\!a \\\"}] ")
     assert caplog.text.count("WARNING") == 0
     caplog.clear()
-    check_integrity(config_2, 2, None, "none",
-                    p_4=[1, 0.5, {
-                        "string": "'[as"
-                    }])
+    check_integrity(config_2, 2, None, "null", p_4=[1, 0.5, {"string": "'[as !a "}])
     with caplog.at_level(logging.WARNING):
         logging.getLogger("yaecs").propagate = True
         mcl(config, config_2.get_command_line_argument(do_return_string=True))
     assert caplog.text.count("WARNING") == 0
     caplog.clear()
-    check_integrity(config, 2, None, "none", p_4=[1, 0.5, {"string": "'[as"}])
+    check_integrity(config_2, 2, None, "null", p_4=[1, 0.5, {"string": "'[as !a "}])
     with caplog.at_level(logging.WARNING):
         logging.getLogger("yaecs").propagate = True
         mcl(config, "--subconfig1.param2")
@@ -295,12 +294,9 @@ def test_merge_from_command_line(caplog, yaml_default, yaml_experiment):
     caplog.clear()
     with caplog.at_level(logging.WARNING):
         logging.getLogger("yaecs").propagate = True
-        mcl(config, "--subconfig1.param2=1")
+        mcl(config, "--subconfig1.param2=yes")
     assert caplog.text.count("WARNING") == 0
     assert config.subconfig1.param2 is True
-    with pytest.raises(Exception, match="could not convert string to float: "
-                       "'False'"):
-        mcl(config, "--param1=False")
 
 
 def test_method_name(caplog):
@@ -361,7 +357,8 @@ def test_variations(capsys, yaml_default):
             }, {
                 "p2": 3.0
             }],
-            "grid": None
+            "grid": None,
+            "tracker_config": {"type": []}
         }, config_class=template(yaml_default), do_not_merge_command_line=True)
     captured = capsys.readouterr()
     assert "WARNING" not in captured.out
@@ -382,6 +379,10 @@ def test_variations(capsys, yaml_default):
     assert (variations[1].p2 == variations[4].p2 == 2.0
             and variations[2].p2 == variations[5].p2 == 3.0)
 
+    def _main(config, tracker):
+        return
+    Experiment(config, _main, experiment_name="test", run_name="test").run(run_description="")
+
 
 def test_pre_processing(capsys, tmp_file_name,
                         yaml_no_file_call_processing_while_loading,
@@ -477,18 +478,19 @@ def test_post_processing(capsys, yaml_default, yaml_experiment, tmp_file_name,
             return f"<Storage: {self.stored}>"
 
     postprocessing = {"*to_store": lambda x: Storage(**x)}
-    config = make_config({
+    default = {
         "a": 10,
         "b.to_store": {
             "i": 1,
             "j": 2
         }
-    }, post_processing_dict=postprocessing)
+    }
+    config = make_config(default, post_processing_dict=postprocessing)
     config.save(str(tmp_file_name))
-    assert config == make_config(str(tmp_file_name),
+    assert config == make_config(default, str(tmp_file_name),
                                  post_processing_dict=postprocessing)
-    assert make_config(str(tmp_file_name)).b.to_store == {"i": 1, "j": 2}
-    assert make_config(str(tmp_file_name)).a == 10
+    assert make_config(default, str(tmp_file_name)).b.to_store == {"i": 1, "j": 2}
+    assert make_config(default, str(tmp_file_name)).a == 10
     # Does post-processing interact correctly with get_command_line_arguments ?
     config = make_config({
         "a": 10,
@@ -504,6 +506,35 @@ def test_post_processing(capsys, yaml_default, yaml_experiment, tmp_file_name,
     assert make_config(dico).a == 10
 
 
+def test_post_processing_modify(yaml_default, yaml_experiment):
+
+    class TestConfig(Configuration):
+        @staticmethod
+        def get_default_config_path():
+            return yaml_default
+
+        def post_processing_func1(self, param1):
+            object.__setattr__(self.get_main_config().subconfig1, "param2", 4)
+            return param1 + 1
+
+        def post_processing_func2(self, param4):
+            object.__setattr__(self.get_main_config().subconfig2, "param3", 5)
+            return param4 + "_"
+
+        def parameters_pre_processing(self):
+            return {
+                "*_path": self.register_as_additional_config_file
+            }
+
+        def parameters_post_processing(self):
+            return {
+                "param1": self.post_processing_func1,
+                "subconfig2.subconfig3.param4": self.post_processing_func2
+            }
+    config = TestConfig.load_config(yaml_experiment, do_not_merge_command_line=True)
+    check_integrity(config, p_1=1.1, p_2=4, p_3=5, p_4="string_")
+
+
 def test_save_reload(capsys, tmp_file_name, yaml_default, yaml_experiment):
     config = load_config(yaml_experiment, default_config=yaml_default)
     captured = capsys.readouterr()
@@ -627,6 +658,73 @@ def test_pattern_matching():
     assert not compare_string_pattern("abcdefgh0123", "*3*3*3")
 
 
+def test_typecheck(yaml_type_check):
+    load_config(default_config=yaml_type_check)
+
+
+def test_yaml_tag_assignment(yaml_tag_assignment_check):
+    template_class = template(default_config=yaml_tag_assignment_check)
+
+    @assign_yaml_tag("add_1", "post", "float")
+    def add_1(self, param):
+        return param + 1
+
+    template_class.add_1 = add_1
+    config = template_class.load_config(do_not_merge_command_line=True)
+    check_integrity(config, p_1=1.1, p_2=3.0, p_3=20.0, p_4=1.1)
+
+
+def test_yaml_order(yaml_default):
+    @assign_order(Priority.SITUATIONAL)
+    def add_1(value):
+        return value + 1
+
+    @assign_order(Priority.OFTEN_FIRST)
+    def double_value_first(value):
+        return value * 2
+
+    @assign_order(Priority.OFTEN_LAST)
+    def double_value_last(value):
+        return value * 2
+    postprocessing = {"param1": add_1, "param1 ": double_value_first}
+    config = load_config(default_config=yaml_default, postprocessing=postprocessing)
+    check_integrity(config, p_1=1.2, p_2=3.0, p_3=20.0)
+    postprocessing = {"param1": add_1, "param1 ": double_value_last}
+    config = load_config(default_config=yaml_default, postprocessing=postprocessing)
+    check_integrity(config, p_1=2.2, p_2=3.0, p_3=20.0)
+
+
+def test_compose(yaml_default):
+    def add_1(value):
+        return value + 1
+
+    def double_value(value):
+        return value * 2
+
+    postprocessing = {"param1": (add_1, double_value)}
+    config = load_config(default_config=yaml_default, postprocessing=postprocessing)
+    check_integrity(config, p_1=2.2, p_2=3.0, p_3=20.0)
+
+
+def test_config_vs_dict_checks(caplog, config_vs_dict_checks):
+
+    with caplog.at_level(logging.WARNING):
+        logging.getLogger("yaecs").propagate = True
+        config = load_config(default_config=config_vs_dict_checks)
+    assert caplog.text.count("WARNING") == 0
+    string = f"\nMAIN CONFIG :\nConfiguration hierarchy :\n> {config_vs_dict_checks}" \
+             "\n\n - config1 : \n	CONFIG1 CONFIG :\n	 - config2 : \n		CONFIG2 CONFIG :\n" \
+             "		 - param1 : {'key1': {'key2': 0}, 'key3': ['!type:dict']}\n\n\n - config3 : \n	CONFIG3 CONFIG :" \
+             "\n	 - config4 : \n		CONFIG4 CONFIG :\n		 - config5 : \n			CONFIG5 CONFIG :" \
+             "\n			 - config6 : \n				CONFIG6 CONFIG :\n				 - param1 : {'key1': {'key1':" \
+             " 0}, 'key2': ['!type:dict']}\n\n\n\n\n - config10 : \n	CONFIG10 CONFIG :\n	 - config11 : " \
+             "\n		CONFIG11 CONFIG :\n		 - config1 : \n			CONFIG1 CONFIG :\n			 - config7 : " \
+             "\n				CONFIG7 CONFIG :\n				 - config8 : \n					CONFIG8 CONFIG :" \
+             "\n					 - param1 : {'key1': {'key1': 0}, 'key2': ['!type:dict']}\n\n				 - co" \
+             "nfig9 : \n					CONFIG9 CONFIG :\n					 - param2 : None\n\n\n\n\n\n"
+    assert config.details() == string
+
+
 def test_warnings(caplog, tmp_file_name):
     # config = ConfigForTests(config_path_or_dictionary={
     #     "param": None, "lparam": [], "dparam": {"param2": 1}})
@@ -655,8 +753,8 @@ def test_warnings(caplog, tmp_file_name):
     assert caplog.text.count("will be ignored : it does not match any") == 1
 
 
-def test_errors(caplog, yaml_default_unlinked, yaml_default_sub_variations,
-                yaml_default_set_twice, yaml_default):
+def test_errors(caplog, yaml_default_sub_variations,
+                yaml_default_set_twice, yaml_default, yaml_type_check):
     caplog.clear()
     with caplog.at_level(logging.WARNING):
         logging.getLogger("yaecs").propagate = True
@@ -711,7 +809,7 @@ def test_errors(caplog, yaml_default_unlinked, yaml_default_sub_variations,
             make_config({"param": 1, "var": {"a": 1}}, config_class=template())
         with pytest.raises(Exception, match="Grid parsing failed.*"):
             make_config({"param": 1, "grid": {}}, config_class=template())
-        with pytest.raises(Exception, match="ERROR : path not found .*"):
+        with pytest.raises(Exception, match="ERROR : no YAML file found at path .*"):
             template()(config_path_or_dictionary="not_found")
         with pytest.raises(Exception, match="'config_metadata' is a "
                                             "special parameter.*"):
@@ -719,7 +817,7 @@ def test_errors(caplog, yaml_default_unlinked, yaml_default_sub_variations,
         with pytest.raises(
                 Exception, match="'overwriting_regime' is a "
                                  "special parameter.*"):
-            metadata = ("Saving time : <date> (<in_seconds>) ; Regime : "
+            metadata = ("Saving time : 0 (0) ; Regime : "
                         "something_incorrect")
             make_config({"config_metadata": metadata})
         with pytest.raises(Exception, match="Failed to set parameter.*"):
@@ -731,10 +829,7 @@ def test_errors(caplog, yaml_default_unlinked, yaml_default_sub_variations,
                 Exception, match=".*character is not authorised "
                                  "in the default config.*"):
             _ = make_config({"param*": 1})
-        with pytest.raises(Exception, match=".*Unlinked sub-configs are not "
-                                            "allowed.*"):
-            _ = make_config(yaml_default_unlinked)
-    assert caplog.text.count("ERROR while pre-processing param") == 4
+    assert caplog.text.count("ERROR while processing param") == 4
     with pytest.raises(
             Exception, match=".*Please declare all your variations "
             "in the main config.*"):
@@ -744,22 +839,52 @@ def test_errors(caplog, yaml_default_unlinked, yaml_default_sub_variations,
             "used as a parameter.*"):
         _ = make_config({"_nesting_hierarchy": 1})
     with pytest.raises(
-            Exception, match=".*cannot be merged : it is not in the"
+            Exception, match=".*cannot be merged : 'subconfig' is not in the"
             " default.*"):
         config = make_config({"param": 1})
         config.merge({"subconfig.param": 1})
     with pytest.raises(
-            Exception, match=".*cannot be merged : it is not in the"
+            Exception, match=".*cannot be merged : 'param2' is not in the"
             " default.*"):
         config = make_config({"param": 1})
         config.merge({"param2": 1})
     with pytest.raises(Exception, match=".*This replacement cannot be "
                        "performed.*"):
-        subconfig = make_config({"param": 1})
-        config = make_config({"subconfig": subconfig})  # don't do this at home
+        config = make_config({"subconfig.param": 1})
         config.merge({"subconfig": 1})
     with pytest.raises(Exception, match=".* was set twice.*"):
         _ = make_config({
             "param": 1,
             "set_twice_path": yaml_default_set_twice
         }, config_class=template())
+    replacements = {
+        "param_int": 1.2,
+        "param_float": "q",
+        "param_str": None,
+        "param_none": [],
+        "param_bool": {},
+        "param_list": True,
+        "param_dict": 5,
+        "param_listint": [0, 2, "3"],
+        "param_listintstr": ["q", 1],
+        "param_dictlistoptionalint": {"a": [None], "b": [1, None, 2.5]}
+    }
+    for k, v in replacements.items():
+        for prefix in ["", "subconfig."]:
+            with pytest.raises(Exception, match=".*has incorrect type '.*"):
+                print(f"Testing for {prefix + k}")
+                c = load_config(default_config=yaml_type_check)
+                c.merge({prefix + k: v})
+
+
+def test_empty_string_in_list(yaml_empty_string_in_list):
+    config = load_config(default_config=yaml_empty_string_in_list)
+    assert config.a == [""]
+
+
+def test_last_mapping_reset_on_new_document(yaml_last_mapping_reset_on_new_document):
+    config = load_config(default_config=yaml_last_mapping_reset_on_new_document)
+    assert config.c_1.p_1_1 == []
+    assert config.c_1.p_1_2 == 0
+    assert config.c_2.p_2_1 == []
+    assert config.c_2.p_2_2 == 0
