@@ -17,30 +17,33 @@ Copyright (C) 2022  Reactive Reality
 """
 
 import logging
-import sys
 import time
-from typing import Callable, Dict, List, Optional, Union
+from typing import Callable, Dict, List, Optional
 
+from ..yaecs_utils import ConfigInput, ConfigDeclarator, format_str, get_config_from_argv, is_config_in_argv
 from .config_base import _ConfigurationBase
-from ..yaecs_utils import ConfigDeclarator, format_str
 
 YAECS_LOGGER = logging.getLogger(__name__)
 
 
 class Configuration(_ConfigurationBase):
-    """ Superclass for YAECS configurations. The superclass implements constructors, while the behaviour is spread
+    """
+    Superclass for YAECS configurations. The superclass implements constructors, while the behaviour is spread
     across the parent classes in the following way :
-    - _ConfigurationBase implements the most basic functionalities such as creation and merging operations. It inherits
+
+    * _ConfigurationBase implements the most basic functionalities such as creation and merging operations. It inherits
       from all other parent classes as Mixins ;
-    - ConfigHooksMixin implements processing functions whose name start with "register_as_" and are decorated by the
+    * ConfigHooksMixin implements processing functions whose name start with `register_as_` and are decorated by the
       yaecs_utils.hook decorator. Users can use those processing either as pre- or post-processing functions, which will
       have the added effect of tagging the processed parameters as playing a certain pre-defined role in the config ;
-    - ConfigGettersMixin implements public getters to access some private attributes as well as other values which
+    * ConfigGettersMixin implements public getters to access some private attributes as well as other values which
       require boilerplate code to be rigorously queried, such as user-defined parameters ;
-    - ConfigSettersMixin implements a few setters which the config uses internally to manipulate private attributes of
+    * ConfigSettersMixin implements a few setters which the config uses internally to manipulate private attributes of
       other Configuration instances (such as sub-configs) ;
-    - ConfigConvenienceMixin implements all functions which are not of central importance to the behaviour of the config
-      but wrap convenient functionalities which can be used either internally of by the user. """
+    * ConfigConvenienceMixin implements all functions which are not of central importance to the behaviour of the config
+      but wrap convenient functionalities which can be used either internally of by the user.
+
+    """
 
     def __init__(self, name: str = "main", overwriting_regime: str = "auto-save",
                  config_path_or_dictionary: Optional[ConfigDeclarator] = None,
@@ -51,13 +54,15 @@ class Configuration(_ConfigurationBase):
         """
         Should never be called directly by the user. Please use one of the constructors instead (load_config,
         build_from_configs, build_from_argv), or the utils.make_config convenience function.
+
         :param name: name for the config or sub-config
         :param overwriting_regime: can be "auto-save" (default, when a param is overwritten it is merged instead and the
-        config is saved automatically if it had been saved previously), "locked" (params can't be overwritten except
-        using merge explicitly) or "unsafe" (params can be freely overwritten but reproducibility is not guaranteed).
+            config is saved automatically if it had been saved previously), "locked" (params can't be overwritten except
+            using merge explicitly) or "unsafe" (params can be freely overwritten but reproducibility
+            is not guaranteed).
         :param config_path_or_dictionary: path or dictionary to create the config from
         :param nesting_hierarchy: list containing the names of all the configs in the sub-config chain leading to this
-        config
+            config
         :param state: processing state used for state tracking and debugging
         :param main_config: main config corresponding to this sub-config, or None if this config is the main config
         :param variation: the name of the variation being created
@@ -68,7 +73,7 @@ class Configuration(_ConfigurationBase):
         :return: none
         """
 
-        # PROTECTED ATTRIBUTES
+        # Set protected attributes
         object.__setattr__(self, "_operating_creation_or_merging", True)
         self._state = [] if state is None else state
         self._main_config = self if main_config is None else main_config
@@ -85,7 +90,7 @@ class Configuration(_ConfigurationBase):
         super().__init__(**kwargs)
         self._protected_attributes = list(self.__dict__) + ["_protected_attributes"]
 
-        # SPECIAL ATTRIBUTES
+        # Set config metadata
         self.config_metadata = {
             "saving_time": time.time(),
             "config_hierarchy": [],
@@ -96,35 +101,44 @@ class Configuration(_ConfigurationBase):
             raise ValueError("'overwriting_regime' needs to be either 'auto-save', "
                              "'locked' or 'unsafe'.")
 
-        # INITIALISATION
+        # Initialise config
         self._state.append(f"setup;{self._name}")
         config_path_or_dictionary = (self.get_default_config_path()
                                      if config_path_or_dictionary is None else config_path_or_dictionary)
         self.init_from_config(config_path_or_dictionary)
-        self.config_metadata["config_hierarchy"] = ([] if not self._nesting_hierarchy else list(
-            main_config.get('.'.join(self._nesting_hierarchy[:-1]), main_config).config_metadata['config_hierarchy']))
+
+        # Find config hierarchy from parents and set it
+        param = None
+        if self._nesting_hierarchy:
+            for index in range(len(self._nesting_hierarchy) - 1):
+                param = main_config.get('.'.join(self._nesting_hierarchy[:index + 1]), main_config)
+                if not isinstance(param, _ConfigurationBase):
+                    param = None  # encountered in temporary configs when merging a deep dict
+                    break
+        self.config_metadata["config_hierarchy"] = [] if param is None else param.config_metadata['config_hierarchy']
         self.config_metadata["config_hierarchy"] += [config_path_or_dictionary]
-        if not self._nesting_hierarchy:
-            self._check_for_unlinked_sub_configs()
+
+        # Checkup and post-setup operations
         if main_config is None:
-            self.set_pre_processing(True)
+            self.get_setter().set_pre_processing(True)
         self._state.pop(-1)
         self._operating_creation_or_merging = False
 
     @classmethod
-    def load_config(cls, *configs: Union[List[ConfigDeclarator],
-                                         ConfigDeclarator], default_config_path: Optional[ConfigDeclarator] = None,
+    def load_config(cls, *configs: ConfigInput, default_config_path: Optional[ConfigDeclarator] = None,
                     overwriting_regime: str = "auto-save", do_not_merge_command_line: bool = False,
                     do_not_pre_process: bool = False, do_not_post_process: bool = False,
                     **kwargs) -> 'Configuration':
         """
         First creates a config using the default config, then merges config_path into it. If config_path is a list,
         successively merges all configs in the list instead from index 0 to the last.
+
         :param configs: config's path or dictionary, or list of default config's paths or dictionaries to merge
         :param default_config_path: default config's path or dictionary
         :param overwriting_regime: can be "auto-save" (default, when a param is overwritten it is merged instead and the
-        config is saved automatically if it had been saved previously), "locked" (params can't be overwritten except
-        using merge explicitly) or "unsafe" (params can be freely overwritten but reproducibility is not guaranteed).
+            config is saved automatically if it had been saved previously), "locked" (params can't be overwritten except
+            using merge explicitly) or "unsafe" (params can be freely overwritten but reproducibility
+            is not guaranteed).
         :param do_not_merge_command_line: if True, does not try to merge the command line parameters
         :param do_not_pre_process: if true, pre-processing is deactivated in this initialization
         :param do_not_post_process: if true, post-processing is deactivated in this initialization
@@ -148,21 +162,22 @@ class Configuration(_ConfigurationBase):
                 config._merge(to_merge, do_not_pre_process=do_not_pre_process, do_not_post_process=do_not_post_process,
                               source="command line")
         config._post_process_modified_parameters()
-        config.set_post_processing(True)
+        config.get_setter().set_post_processing(True)
         return config
 
     @classmethod
-    def build_from_configs(cls, *configs: Union[List[ConfigDeclarator], ConfigDeclarator],
-                           overwriting_regime: str = "auto-save", do_not_merge_command_line: bool = False,
-                           do_not_pre_process: bool = False, do_not_post_process: bool = False,
-                           **kwargs) -> 'Configuration':
+    def build_from_configs(cls, *configs: ConfigInput, overwriting_regime: str = "auto-save",
+                           do_not_merge_command_line: bool = False, do_not_pre_process: bool = False,
+                           do_not_post_process: bool = False, **kwargs) -> 'Configuration':
         """
         First creates a config using the first config provided (or the first config in the provided list), then merges
         the subsequent configs into it from index 1 to the last.
+
         :param configs: config's path or dictionary, or list of default config's paths or dictionaries to merge
         :param overwriting_regime: can be "auto-save" (default, when a param is overwritten it is merged instead and the
-        config is saved automatically if it had been saved previously), "locked" (params can't be overwritten except
-        using merge explicitly) or "unsafe" (params can be freely overwritten but reproducibility is not guaranteed).
+            config is saved automatically if it had been saved previously), "locked" (params can't be overwritten except
+            using merge explicitly) or "unsafe" (params can be freely overwritten but reproducibility
+            is not guaranteed).
         :param do_not_merge_command_line: if True, does not try to merge the command line parameters
         :param do_not_pre_process: if true, pre-processing is deactivated in this initialization
         :param do_not_post_process: if true, post-processing is deactivated in this initialization
@@ -184,7 +199,7 @@ class Configuration(_ConfigurationBase):
                                do_not_pre_process=do_not_pre_process, do_not_post_process=do_not_post_process, **kwargs)
 
     @classmethod
-    def build_from_argv(cls, *configs: Union[List[ConfigDeclarator], ConfigDeclarator], fallback: Optional[str] = None,
+    def build_from_argv(cls, *configs: ConfigInput, fallback: Optional[ConfigInput] = None, pattern: str = "--config",
                         default_config_path: Optional[ConfigDeclarator] = None, overwriting_regime: str = "auto-save",
                         do_not_merge_command_line: bool = False, do_not_pre_process: bool = False,
                         do_not_post_process: bool = False, **kwargs) -> 'Configuration':
@@ -192,14 +207,17 @@ class Configuration(_ConfigurationBase):
         Assumes a pattern of the form '--config <path_to_config>' or '--config [<path1>,<path2>,...]' in sys.argv (the
         brackets are optional), and builds a config from the specified paths by merging them into the default config in
         the specified order.
+
         :param configs: config's path or dictionary, or list of config paths or dictionaries to merge. Those will be
-        merged to the default config before the config from the command line.
+            merged to the default config before the config from the command line.
         :param fallback: config path or dictionary, or list of config paths or dictionaries to fall back to if no config
-        was detected in the argv
+            was detected in the argv
+        :param pattern: pattern to look for in sys.argv
         :param default_config_path: default config's path or dictionary
         :param overwriting_regime: can be "auto-save" (default, when a param is overwritten it is merged instead and the
-        config is saved automatically if it had been saved previously), "locked" (params can't be overwritten except
-        using merge explicitly) or "unsafe" (params can be freely overwritten but reproducibility is not guaranteed).
+            config is saved automatically if it had been saved previously), "locked" (params can't be overwritten except
+            using merge explicitly) or "unsafe" (params can be freely overwritten but reproducibility
+            is not guaranteed).
         :param do_not_merge_command_line: if True, does not try to merge the command line parameters
         :param do_not_pre_process: if true, pre-processing is deactivated in this initialization
         :param do_not_post_process: if true, post-processing is deactivated in this initialization
@@ -207,27 +225,25 @@ class Configuration(_ConfigurationBase):
         :raises TypeError: if --config is not found and no fallback
         :return: instance of Configuration object containing the desired config
         """
-        if "--config" not in sys.argv and fallback is None:
-            raise TypeError("The pattern '--config' was not detected "
-                            "in sys.argv.")
-        if "--config" in sys.argv:
-            fallback = [cfg.strip(" ") for cfg in sys.argv[sys.argv.index("--config") + 1].strip("[]").split(",")]
-        if not isinstance(fallback, list):
-            fallback = [fallback]
+        configs_from_argv: List[ConfigDeclarator] = get_config_from_argv(pattern=pattern, fallback=fallback)
 
-        return cls.load_config(*configs, *fallback, default_config_path=default_config_path,
+        return cls.load_config(*configs, *configs_from_argv, default_config_path=default_config_path,
                                overwriting_regime=overwriting_regime,
                                do_not_merge_command_line=do_not_merge_command_line,
                                do_not_pre_process=do_not_pre_process, do_not_post_process=do_not_post_process,
-                               from_argv=True, **kwargs)
+                               from_argv=pattern if is_config_in_argv(pattern=pattern) else "", **kwargs)
 
     def create_variations(self) -> List['Configuration']:
         """
         Creates a list of configs that are derived from the current config using the internally tracked variations and
         grids registered via the corresponding functions (register_as_config_variations and register_as_grid).
+
         :raises TypeError: if grid dimension is empty or not registered
         :return: the list of configs corresponding to the tracked variations
         """
+        if self._variation_name is not None:
+            return [self]  # if this config is already a variation, it should not create further variations
+
         variations_names_to_use = [i[0] for i in self._configuration_variations]
         variations_names_to_use_changing = [i[0] for i in self._configuration_variations]
         variations_to_use = [i[1] for i in self._configuration_variations]
@@ -282,6 +298,8 @@ class Configuration(_ConfigurationBase):
                         variations_names.append(name + "_" + var_name[1][variation_index])
 
         # Creating configs
+        if not variations:
+            return [self]
         variation_configs = []
         for variation_index, variation in enumerate(variations):
             variation_configs.append(
@@ -291,9 +309,7 @@ class Configuration(_ConfigurationBase):
                                            do_not_merge_command_line=True, variation=variations_names[variation_index],
                                            verbose=False))
             variation_configs[-1].set_variation_name(variations_names[variation_index], deep=True)
-        if not variation_configs:
-            variation_configs = [self]
-        elif self._verbose:
+        if self._verbose:
             YAECS_LOGGER.info(f"Created {len(variation_configs)} variation{'s' if len(variation_configs) > 1 else ''} "
                               f"from registered variation parameters.")
         return variation_configs
@@ -302,6 +318,7 @@ class Configuration(_ConfigurationBase):
     def get_default_config_path() -> str:
         """
         Returns the path to the default config of the project. This function must be implemented at project-level.
+
         :return: string corresponding to the path to the default config of the project
         """
         raise NotImplementedError
@@ -348,13 +365,14 @@ class Configuration(_ConfigurationBase):
     def set_variation_name(self, name: str, deep: bool = False) -> None:
         """
         Sets the variation index of the config. This function is not intended to be used by the user.
+
         :param name: index to set the variation index with
         :param deep: whether to also recursively set the variation name of all sub-configs
         """
         object.__setattr__(self, "_variation_name", name)
         if deep:
-            for subconfig in self._sub_configs_list:
-                subconfig.set_variation_name(name, deep=True)
+            for subconfig in self.get_sub_configs(deep=True):
+                object.__setattr__(subconfig, "_variation_name", name)
 
     @classmethod
     def _get_instance(cls, name: str = "main", overwriting_regime: str = "auto-save",
