@@ -16,17 +16,19 @@ Copyright (C) 2022  Reactive Reality
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 """
 
-from copy import deepcopy
 import difflib
-from functools import partial
 import logging
 import os
 import time
-from typing import (Any, Callable, Dict, ItemsView, KeysView, List, Optional, Tuple, Type, TYPE_CHECKING, Union,
-                    ValuesView)
+from copy import deepcopy
+from functools import partial
+from typing import (TYPE_CHECKING,
+                    Any, Callable, Dict, ItemsView, KeysView, List, Optional, Tuple, Type, Union, ValuesView)
+
 import yaml
 
 from ..yaecs_utils import compare_string_pattern, dict_apply, format_str
+
 if TYPE_CHECKING:
     from .config import Configuration
 
@@ -52,6 +54,9 @@ class ConfigConvenienceMixin:
     _verbose: bool
     _was_last_saved_as: Optional[str]
 
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+
     def __eq__(self, other: Any) -> bool:
         if not isinstance(other, ConfigConvenienceMixin):
             return False
@@ -69,7 +74,7 @@ class ConfigConvenienceMixin:
         return True
 
     def __hash__(self) -> int:
-        return hash(repr(self.get_dict(True)))
+        return hash(repr(self.get_dict(deep=True, pre_post_processing_values=False)))
 
     def __repr__(self) -> str:
         return "<Configuration:" + self.get_name() + ">"
@@ -80,6 +85,7 @@ class ConfigConvenienceMixin:
         configuration and the "other" configuration. Tuples are written in the form :
         (parameter_name, parameter_value_in_other). If parameter_name does not exist in other, (parameter_name, None) is
         given instead.
+
         :param other: config to compare self with
         :param reduce: tries to reduce the size of the output text as much as possible
         :return: difference list
@@ -140,28 +146,31 @@ class ConfigConvenienceMixin:
     def copy(self) -> 'ConfigConvenienceMixin':
         """
         Returns a safe, independent copy of the config
+
         :return: instance of Configuration that is a deep copy of the config
         """
         return deepcopy(self)
 
-    def details(self, show_only: Optional[Union[str, List[str]]] = None,
+    def details(self, shorten: int = -1, show_only: Optional[Union[str, List[str]]] = None,
                 expand_only: Optional[Union[str, List[str]]] = None, no_show: Optional[Union[str, List[str]]] = None,
                 no_expand: Optional[Union[str, List[str]]] = None) -> str:
         """
         Creates and returns a string describing all the parameters in the config and its sub-configs.
+
+        :param shorten: if > 0, params are shortened to that many characters except if logging level is under INFO.
         :param show_only: if not None, list of names referring to params. Only params in the list are displayed in the
-        details.
+            details.
         :param expand_only: if not None, list of names referring to sub-configs. Only sub-configs in the list are
-        unrolled in the details.
+            unrolled in the details.
         :param no_show: if not None, list of names referring to params. Params in the list are not displayed in the
-        details.
+            details.
         :param no_expand: if not None, list of names referring to sub-configs. Sub-configs in the list are not unrolled
-        in the details.
+            in the details.
         :return: string containing the details
         """
         constraints = {"show_only": show_only, "expand_only": expand_only, "no_show": no_show, "no_expand": no_expand}
         constraints = dict_apply(constraints, self.match_params)
-        string_to_return = ("\n" + "\t" * len(self._nesting_hierarchy) + self.get_name().upper() + " CONFIG :\n")
+        string_to_return = "\n" + "\t" * len(self._nesting_hierarchy) + self.get_name().upper() + " CONFIG :\n"
         if not self._nesting_hierarchy:
             string_to_return += "Configuration hierarchy :\n"
             for hierarchy_level in self.config_metadata["config_hierarchy"]:
@@ -186,43 +195,49 @@ class ConfigConvenienceMixin:
                         and (constraints["expand_only"] is None
                              or attribute in [cstr.split(".")[0] for cstr in constraints["expand_only"]])):
                     _for_sub_config_attr = partial(_for_sub_config, attribute=attribute)
-                    string_to_return += (self[attribute].details(**(dict_apply(constraints, _for_sub_config_attr)))
-                                         + "\n")
+                    string_to_return += self[attribute].details(shorten=shorten,
+                                                                **(dict_apply(constraints, _for_sub_config_attr)))
                 else:
                     string_to_return += self[attribute].get_name().upper()
-                    string_to_return += "\n"
             else:
-                string_to_return += str(self[attribute]) + "\n"
+                string_to_return += format_str(str(self[attribute]), shorten) if shorten > 0 else str(self[attribute])
+            string_to_return += "\n"
         return string_to_return
 
-    def items(self, deep: bool = False) -> ItemsView:
+    def items(self, deep: bool = False, pre_post_processing_values: bool = False) -> ItemsView:
         """
         Behaves as dict.items(). If deep is False, sub-configs remain sub-configs in the items. Otherwise, they are
-        converted to dict.
+        converted to dict. If pre_post_processing_values is True, the values are returned in the state they had before
+        the post-processing step.
+
         :param deep: how to return sub-configs that would appear among the items. If False, do not convert them, else
-        recursively convert them to dict
+            recursively convert them to dict
+        :param pre_post_processing_values: whether to return the values in the state they had before the post-processing
+            step
         :return: the items of the config as in dict.items()
         """
-        return self.get_dict(deep).items()
+        return self.get_dict(deep=deep, pre_post_processing_values=pre_post_processing_values).items()
 
     def keys(self) -> KeysView:
         """
         Behaves as dict.keys(), returning a _dict_keys instance containing the names of the params of the config.
+
         :return: the keys if the config as in dict.keys()
         """
-        return self.get_dict(False).keys()
+        return self.get_dict(deep=False, pre_post_processing_values=False).keys()
 
     def match_params(self, *patterns: Optional[Union[str, List[str]]]) -> Optional[List[str]]:
         """
         For a string, a list of strings or several strings, returns all params matching at least one of the input
         strings.
+
         :param patterns: string, a list of strings or several strings
         :return: all params matching at least one of the input string
         """
         patterns = (patterns[0] if len(patterns) == 1 and isinstance(patterns[0], list) else patterns)
         if patterns is None or (len(patterns) == 1 and patterns[0] is None):
             return None
-        new_names = [n for n in patterns if "*" not in n and n in self]
+        new_names = [n.strip(" ") for n in patterns if "*" not in n and n.strip(" ") in self.get_parameter_names(True)]
         for name in [n for n in patterns if "*" in n]:
             new_names = new_names + [p for p in self.get_parameter_names(True) if compare_string_pattern(p, name)]
         return new_names
@@ -230,11 +245,12 @@ class ConfigConvenienceMixin:
     def save(self, filename: str = None, save_header: bool = True, save_hierarchy: str = True) -> None:
         """
         Saves the current config at the provided location. The saving format allows for a perfect recovery of the config
-        by using : config = Configuration.load_config(filename). If no filename is given, overwrites the last save.
+        by using : `config = Configuration.load_config(filename)`. If no filename is given, overwrites the last save.
+
         :param filename: path to the saving location of the config
         :param save_header: whether to save the config metadata as the fist parameter. This will tag the saved file as a
-        saved config in the eye of the config system when it gets merged, which will deactivate pre-processing
-        :param save_hierarchy: whether to save config hierarchy as a '*_hierarchy.yaml' file
+            saved config in the eye of the config system when it gets merged, which will deactivate pre-processing
+        :param save_hierarchy: whether to save config hierarchy as a `*_hierarchy.yaml` file
         :raises RuntimeError: no file name is provided and there is no previous save to overwrite
         """
         if filename is None:
@@ -248,8 +264,8 @@ class ConfigConvenienceMixin:
         config_dump_path = file_path + file_extension
         to_dump = {
             a: (getattr(self, "___" + a if a in self._methods else a) if
-                (self._get_full_path(a) not in self.get_main_config().get_pre_post_processing_values()) else
-                self.get_main_config().get_pre_post_processing_values()[self._get_full_path(a)]
+                (self._get_full_path(a) not in self._main_config.get_pre_post_processing_values()) else
+                self._main_config.get_pre_post_processing_values()[self._get_full_path(a)]
                 ) if a != "config_metadata" else self._format_metadata()
             for a in (["config_metadata"] if save_header else []) + self._get_user_defined_attributes()
         }
@@ -260,21 +276,41 @@ class ConfigConvenienceMixin:
             hierarchy_dump_path = f"{file_path}_hierarchy{file_extension}"
             to_dump = {"config_hierarchy": self.config_metadata["config_hierarchy"]}
             with open(hierarchy_dump_path, "w", encoding='utf-8') as fil:
-                yaml.dump(to_dump, fil, Dumper=self._get_yaml_dumper(), width=1000)
+                yaml.dump(to_dump, fil, Dumper=self._get_yaml_dumper(dict_behaviour="default"), width=1000)
 
         object.__setattr__(self, "_was_last_saved_as", config_dump_path)
         if self._verbose:
             YAECS_LOGGER.info(f"Configuration saved in : {format_str(os.path.abspath(config_dump_path))}.")
 
-    def values(self, deep: bool = False) -> ValuesView:
+    def values(self, deep: bool = False, pre_post_processing_values: bool = False) -> ValuesView:
         """
         Behaves as dict.values(). If deep is False, sub-configs remain sub-configs in the values. Otherwise, they are
-        converted to dict.
+        converted to dict. If pre_post_processing_values is True, the values are returned in the state they had before
+        the post-processing step.
+
         :param deep: how to return sub-configs that would appear among the values. If False, do not convert them, else
-        recursively convert them to dict
+            recursively convert them to dict
+        :param pre_post_processing_values: whether to return the values in the state they had before the post-processing
+            step
         :return: the values of the config as in dict.values()
         """
-        return self.get_dict(deep).values()
+        return self.get_dict(deep=deep, pre_post_processing_values=pre_post_processing_values).values()
+
+    @staticmethod
+    def _are_same_sub_configs(first: Any, second: Any) -> bool:
+        """
+        Checks if two sub-configs have identical nesting hierarchies.
+
+        :param first: first sub-config to check
+        :param second: second sub-config to check
+        :return: result of the check
+        """
+        if not isinstance(first, ConfigConvenienceMixin) or not isinstance(second, ConfigConvenienceMixin):
+            return False
+        if first.get_name() != second.get_name():
+            return False
+        nh1, nh2 = first.get_nesting_hierarchy(), second.get_nesting_hierarchy()
+        return len(nh1) == len(nh2) and all(nh1[i] == nh2[i] for i in range(len(nh1)))
 
     def _did_you_mean(self, name: str, filter_type: Optional[type] = None, suffix: str = "") -> str:
         """ Used to propose suggestions when the user tries to access a parameter which does not exist. """
@@ -301,26 +337,37 @@ class ConfigConvenienceMixin:
         """ Used to format the metadata for saving or printing. """
         return (f"Saving time : {time.ctime(self.config_metadata['saving_time'])} "
                 f"({self.config_metadata['saving_time']}) ; "
-                f"Regime : {self.config_metadata['overwriting_regime']}")
+                f"Regime : {self.config_metadata['overwriting_regime']}"
+                f"{'' if self.get_variation_name() is None else f' ; Variation : {self.get_variation_name()}'}")
 
-    def _get_yaml_dumper(self) -> Type[yaml.Dumper]:
+    def _get_yaml_dumper(self, dict_behaviour: str = "custom") -> Type[yaml.Dumper]:
         """ Used to get a custom YAML dumper capable of writing config tags. """
 
         def config_representer(yaml_dumper, class_instance):
             # pylint: disable=bad-continuation
             return yaml_dumper.represent_mapping(
-                "!" + class_instance.get_name().split("_VARIATION_")[0], {
+                "tag:yaml.org,2002:map", {
                     a[3:] if a.startswith("___") else a: self._format_metadata() if a == "config_metadata" else
                     (b if (".".join(class_instance.get_nesting_hierarchy()
-                                    + [a]) not in self.get_main_config().get_pre_post_processing_values()) else
-                        (self.get_main_config().get_pre_post_processing_values(
-                             )[".".join(class_instance.get_nesting_hierarchy() + [a])]))
+                                    + [a]) not in self._main_config.get_pre_post_processing_values()) else
+                        (self._main_config.get_pre_post_processing_values(
+                        )[".".join(class_instance.get_nesting_hierarchy() + [a])]))
                     for (a, b) in class_instance.__dict__.items()
                     if a not in self._protected_attributes
                     and not (class_instance.get_nesting_hierarchy() and a in ["config_metadata"])
                 },
             )
 
+        def dict_representer(yaml_dumper, data):
+            if yaml_dumper.represented_objects and dict_behaviour == "custom":
+                return yaml_dumper.represent_mapping('!type:dict', data)
+            return yaml_dumper.represent_mapping('tag:yaml.org,2002:map', data)
+
         dumper = yaml.Dumper
+        dumper.add_representer(dict, dict_representer)
         dumper.add_representer(self.__class__, config_representer)
         return dumper
+
+    def _is_main_config(self) -> bool:
+        """ Returns whether the config is the main config. """
+        return not self._nesting_hierarchy
